#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <stdlib.h>

/***************************************************************************/
/* This program will produce a sorted directory listing.  It gives some    */
/* information that is not provided by DOS's dir.                          */
/*                                                                         */
/* Attribute                                                               */
/*   "r"=ReadOnly, "a"=Archived, "h"=Hidden, "s"=System                    */
/*                                                                         */
/* Disk Stats                                                              */
/*   Amount of disk space used, free, and total on the drive               */
/*                                                                         */
/* File Allocation                                                         */
/*   How much disk space is really allocated to a file.  This is computed  */
/*   by getting from DOS the cluster size and computing how many cluseters */
/*   are needed for the file.  A total of what DOS reports and what is     */
/*   computed is reported at the end as "Excess Allocated"                 */
/***************************************************************************/

int     print_heading_02(char);  /* returns bytes/cluster */
char    *get_sysdate();
char    *get_systime();
char    *get_volid(char);
void    print_heading(char);
char    *remove_period(char *, char *);
char    *get_filedate(int);
char    *get_filetime(int);
void    get_file_ext(char *, char *, char *);
int     dir_compare(struct my_dir *, struct my_dir *);
char    *get_attribute(char);
void    fill_zero(char *);
#define A_ALL_FILE _A_NORMAL | _A_RDONLY | _A_HIDDEN | _A_SYSTEM | _A_ARCH | _A_SUBDIR

struct my_dir
{   char    name[13];
    long    size;
    int     date,
            time;
    char    attr;
};

char    *date_format = "%2d/%2d/%2d";
char    *time_format = "%2d:%2d";
/*                        12345678901234567890123456789012345678901234567
/*                        DI vn.n  VolId: xxxxxxxxxxxx  Date:99/99/99  Time:99:99 */
/*                        xxxxxxxx  xxx  999999999  99/99/99  99:99  xxxx
/*                        xxxxxxxxxxx   <SUBDIR>    99/99/99  99:99  xxxx */
char    *heading_01 = "DI v1.2  VolId: %-12s  Date:%8s  Time: %5s\n";
char    *heading_02 = " Disk  Total: %-8ld    Used: %-8ld   Free: %-8ld\n";
char    *detail_fmt = "%-8s  %-3s  %8ld  %s  %s  %4s %8ld\n";
char    *subdir_fmt = "%-11s   <SubDir>   %s  %s  %s\n";
main (argc, argv)
    int     argc;
    char    *argv[];
{

    struct find_t   dir;
    char            work_string[40], drive_letter, name[13];
    char            fname[9], fext[4];
    int             length, file_count, bytes_per_cluster, bpc_1;
    long int        total_size, total_allocated, allocated;
    char            *temp_ptr;
    register int    i;

#define CMDLEN  80
#define PATHLEN 64
#define CWDLEN  64
    char    command[CMDLEN+1], path[PATHLEN+1], cwd[CWDLEN+1];

/***************************************************************************/
/* set up a table to hold the filespecs for sorting ************************/
/***************************************************************************/
#define DIRMAX 500
    struct my_dir   *mdir, *mdir_ptr1, *mdir_bottom;
    int             dir_count, dir_bottom;
    mdir = (struct my_dir *) calloc(DIRMAX, sizeof(struct my_dir));
    if (mdir == NULL)
    {   puts ("Not enough memory for the table !");
        exit(1);
    }

/***************************************************************************/
/* Begin building the runtime specs ****************************************/
/***************************************************************************/
    getcwd(cwd, CWDLEN);                       /* where we are now CWD     */
    if (argc >1) {
        strcpy(command, argv[1]);
        strupr(command);
    }
    else strcpy(command,cwd);                  /* use command line or cwd ?*/

    if (strchr(command, ':'))                  /* extract drive letter     */
    {   drive_letter = command[0];             /* turn into DOS's format   */
        strcpy(command, command+2);            /* throw away drive letter  */
        cwd[0] = '\0';                         /* "kill" the other         */
    }
    else
    {   drive_letter = cwd[0];                 /* same for the default     */
        strcpy(cwd, cwd+2);
    }

/***************************************************************************/
/* get the path and file name from the command line if possible ************/
/***************************************************************************/
    name[0]='\0';
    if (!strchr(command, '.') && command[strlen(command)-1] != '\\')
        strcat(command, "\\");                 /* tack on "\" if needed    */
    if (strchr(command, '.'))                  /* file or path - add "\"   */
    {   temp_ptr = strrchr(command, '\\');     /* get file name if there   */
        if (! temp_ptr)                        /* no path means all name   */
        {   strcpy(name, command);
            command[0] = '\0';                 /* got all we can get here  */
        }
        else
        {   strcpy(name, temp_ptr+1);          /* start after last "\"     */
            *(temp_ptr+1) = '\0';              /* chop off the name        */
        }
    }
    if (name[0]=='\0') strcpy(name, "*.*");    /* no name then *.*         */

    temp_ptr = strrchr(command, '\\');         /* no go after the path part*/
    if (temp_ptr) strcpy(path, command);       /* \ indicates a path there */
        else strcpy(path, cwd);                /* else use default (above) */
    length = strlen(path);                     /* something not end w/ "\" */
    if (length && path[length-1] != '\\') strcat(path, "\\");

/***************************************************************************/
/* print the program heading line ******************************************/
/***************************************************************************/
    printf(heading_01, get_volid(drive_letter), get_sysdate(),
        get_systime());
    bytes_per_cluster = print_heading_02(drive_letter);
    bpc_1 = bytes_per_cluster-1;

/***************************************************************************/
/* build the dos findfirst/findnext buffer and do DOS **********************/
/***************************************************************************/
    work_string[0] = drive_letter;             /* start with drive         */
    work_string[1] = ':';                      /* add ":"                  */
    work_string[2]='\0';                       /* and prepare for concat   */
    strcat(work_string, path);                 /* add path spec            */
    printf(" Dir of %s\n", work_string);       /* and print the line       */
    strcat(work_string, name);                 /* tack on file's name      */
    mdir_bottom = mdir;                        /* set array address pointer*/
    mdir_bottom--;                             /* start with "-1"          */
    dir_count = 0;                             /* set count to zero        */
    if (_dos_findfirst(work_string, A_ALL_FILE, &dir)) /* do the first     */
    {   puts("No Files");                      /* nothing found !          */
        exit(1);                               /* stop here                */
    }
    else                                       /* something found          */
    {   if(dir.name[0] != '.')                 /* find "." in name.ext     */
        {   dir_count++;                       /* bump the counter         */
            mdir_bottom++;                     /* bump the array pointer   */
            strcpy(mdir_bottom->name, dir.name);   /* add to table         */
            mdir_bottom->date = dir.wr_date;   /* add...                   */
            mdir_bottom->time = dir.wr_time;   /* add...                   */
            mdir_bottom->size = dir.size;      /* add...                   */
            mdir_bottom->attr = dir.attrib;    /* add...                   */
        }
/* read all there is to read ***********************************************/
        while (! _dos_findnext(&dir))          /* DOS call continued...    */
        {   if(dir.name[0] != '.')             /* same general logic as    */
            {   mdir_bottom++;                 /* above - this could become*/
                dir_count++;                   /* a procedure              */
                strcpy(mdir_bottom->name, dir.name);
                mdir_bottom->date = dir.wr_date;
                mdir_bottom->time = dir.wr_time;
                mdir_bottom->size = dir.size;
                mdir_bottom->attr = dir.attrib;
            }
        }
    }

/***************************************************************************/
/* sort the table created above ********************************************/
/***************************************************************************/
    qsort((void *) mdir, (size_t) dir_count, sizeof(struct my_dir),
            (*dir_compare));

/***************************************************************************/
/* process subdirs here in the listing here ********************************/
/***************************************************************************/
    mdir_ptr1 = mdir;                          /* start at the beginning   */
    for (i=0; i<dir_count; i++, mdir_ptr1++)   /* for each entry in array  */
    {   if (mdir_ptr1->attr & _A_SUBDIR)       /* is it a SubDirectory?    */
        {   remove_period(name, mdir_ptr1->name);  /* yes - format for     */
            printf(subdir_fmt, name,               /* output               */
                    get_filedate(mdir_ptr1->date),
                    get_filetime(mdir_ptr1->time),
                    get_attribute(mdir_ptr1->attr));
        }
    }

/***************************************************************************/
/* process real files here *************************************************/
/***************************************************************************/
    mdir_ptr1 = mdir;                          /* start at the beginning   */
    total_size = 0;                            /* clear size totals        */
    total_allocated = 0;
    file_count = 0;                            /* clear file count         */
    for (i=0; i<dir_count; i++, mdir_ptr1++)   /* for each entry in array  */
    {   if (! (mdir_ptr1->attr & _A_SUBDIR))   /* is this a file's entry?  */
        {   get_file_ext(fname, fext, mdir_ptr1->name);            /* yes  */
            allocated = (mdir_ptr1->size + bpc_1)/bytes_per_cluster;
            allocated = allocated * bytes_per_cluster;
            printf(detail_fmt, fname, fext, mdir_ptr1->size,
                    get_filedate(mdir_ptr1->date),
                    get_filetime(mdir_ptr1->time),
                    get_attribute(mdir_ptr1->attr),
                    allocated);
            total_size += mdir_ptr1->size;
            total_allocated += allocated;
            file_count++;
        }
    }

/***************************************************************************/
/*  print end of run statistics ********************************************/
/***************************************************************************/
    printf ("  Files: %-4d  %8ld    Bytes    Allocated: %8ld\n",
                file_count, total_size, total_allocated);
    printf ("    Excess Allocted: %8ld\n", total_allocated-total_size);
} /* end of main block */
/***************************************************************************/

/* FUNTIONS FOLLOW */

/***************************************************************************/
/* replace all spaces with zeros in the string *****************************/
/***************************************************************************/
void fill_zero(str_in)
    char        *str_in;
{
   _asm {
                push        si                 ; save the register
                mov         si,str_in          ; load address given by C
next_si:        inc         si                 ; bump the pointer
                mov         al,[si]            ; let's look at the character
                cmp         al,00h             ; null terminator ?
                je          end_fill_zero      ; yes - all done
                cmp         al,' '             ; is the character a space?
                jne         next_si            ; no  - ignore it
                mov         byte ptr [si],'0'  ; else- set it to a "0"
                jmp short   next_si            ; continue...
end_fill_zero:  pop         si                 ; restore the saved register
        }
}
/***************************************************************************/
/* smaller and faster than C's split_path function *************************/
/***************************************************************************/
char *remove_period(str_out, str_in)           /* this function "squeezes" */
    char        *str_out, *str_in;             /* out any "." in the souce */
{   char        *enter;                        /* string                   */
    enter = str_out;                           /* save the start address   */
    while(*str_in)                             /* look at each before "."  */
    {   if(*str_in != '.') *str_out++ = *str_in;  /* move to all but "."   */
        str_in++;                              /* next source byte         */
    }
    *str_out = '\0';                           /* tack on a new null       */
    return(enter);
}

/***************************************************************************/
/* this function converts DOS's internal date to humanoid form *************/
/***************************************************************************/
char *get_filedate(date_in)
    int     date_in;
{   union                                      /* redefine the bits here   */
    {   int     u_date;
        struct
        {   unsigned int day:5;
            unsigned int month:4;
            unsigned int year:7;
        } dd;
    } d;
    static char fdate[9];                      /* this gets returned       */

    d.u_date = date_in;                        /* move date to redefine    */
    sprintf (fdate, date_format, d.dd.month, d.dd.day, (d.dd.year+80));
    fill_zero(fdate);
    return(fdate);
}
/***************************************************************************/
/* this function converts DOS's internal tile to humaniod form *************/
/***************************************************************************/
char *get_filetime(time_in)
    int     time_in;
{   union                                      /* redefine the bits here   */
    {   int     u_time;
        struct
        {   unsigned int second:5;
            unsigned int minute:6;
            unsigned int hour:5;
        } dd;
    } d;
    static char ftime[9];                      /* this gets returned       */

    d.u_time = time_in;                        /* move time to redefine    */
    sprintf (ftime, time_format, d.dd.hour, d.dd.minute);
    fill_zero(ftime);
    return(ftime);
}

/***************************************************************************/
/* extract the extention from a file name - faster than split_path *********/
/***************************************************************************/
void get_file_ext(fname, fext, filename)
    char        *fname, *fext, *filename;
{   char    *work;
    work = strchr(filename, '.');              /* find "." if there        */
    if (work == NULL)                          /* if not there             */
    {   strcpy(fname, filename);               /*   all name               */
        fext[0]='\0';                          /*   no extention           */
    }
    else                                       /* if there is a "."        */
    {   *work = '\0';                          /* make it the end          */
        strcpy(fname, filename);               /*   all before is name     */
        strcpy(fext, (work+1));                /*   all after is extention */
    }
}

/***************************************************************************/
/* this is the function used by qsort for comparisons **********************/
/***************************************************************************/
int dir_compare(ptr1, ptr2)
    struct my_dir *ptr1, *ptr2;
{
    return (strcmp(ptr1->name, ptr2->name));   /* comp names return result */
}

/***************************************************************************/
/* build a string for output in the listing ********************************/
/***************************************************************************/
char *get_attribute(attrib)
    char    attrib;
{   static char attr_string[5];                /* this gets returned       */
    strcpy(attr_string,"    ");
    if(attrib & '\x20') attr_string[0] = 'a';
    if(attrib & '\x01') attr_string[1] = 'r';
    if(attrib & '\x02') attr_string[2] = 'h';
    if(attrib & '\x04') attr_string[3] = 's';

    return(attr_string);
}

/***************************************************************************/
/* get the volume id from the selected drive *******************************/
/***************************************************************************/
char *get_volid(drive_letter)
    char        drive_letter;
{   static char     vol_id[13];
    static char     *none = "-No Vol Id-";
    struct find_t   buffer;
    char            *vol_id_spec = "?:\\*.*";
    vol_id_spec[0] = drive_letter;

    if (_dos_findfirst(vol_id_spec, _A_VOLID, &buffer) == 0)
    {       strcpy(vol_id, buffer.name);
            remove_period(vol_id, vol_id);
            return(vol_id);
    }
    return(none);
}

/***************************************************************************/
/* get the current date from the system ************************************/
/***************************************************************************/
char *get_sysdate()
{   static char fdate[9];                      /* this gets returned       */
    struct dosdate_t sys_date;                 /* C builtin structure      */

    _dos_getdate(&sys_date);                   /* get date, format outpt   */
    sprintf(fdate, date_format, sys_date.month, sys_date.day,
        sys_date.year-1900);
    fill_zero(fdate);
    return(fdate);
}

/***************************************************************************/
/* get the current time from the system ************************************/
/***************************************************************************/
char *get_systime()
{   static char ftime[6];                      /* this gets returned       */
    struct dostime_t sys_time;                 /* C builtin structure      */

    _dos_gettime(&sys_time);                   /* get time, format output  */
    sprintf(ftime, time_format, sys_time.hour, sys_time.minute);
    fill_zero(ftime);
    return(ftime);
}

/***************************************************************************/
/* this routine calculates the disk usage amounts and prints it ************/
/***************************************************************************/
int print_heading_02(drive_letter)
    char        drive_letter;
{
    struct diskfree_t diskinfo;
    long int        disk_total, disk_used, disk_free;
    int             bytes_per_cluster;
    unsigned int    drive_number;
    drive_number = (unsigned int) drive_letter & '\x0f';

    _dos_getdiskfree(drive_number, &diskinfo);
    bytes_per_cluster = diskinfo.bytes_per_sector * diskinfo.sectors_per_cluster;
    disk_total = (long int) bytes_per_cluster * diskinfo.total_clusters;
    disk_free = (long int) bytes_per_cluster * diskinfo.avail_clusters;
    disk_used = disk_total - disk_free;
    printf(heading_02, disk_total, disk_used, disk_free);
    return(bytes_per_cluster);
}
/***************************************************************************/
/***************************************************************************/
